///////////////////////////////////////////////////
// PerfSleep.cpp
// Part of 720020-1152x PerfSleep
//
// Applied Data Systems
//
// See PerfSleep.h for history and notes.
//
///////////////////////////////////////////////////

#include "PerfSleep.h"

#ifdef PXA255
	#define OST_BASE			0x40A00000  // OS Timer
	#define OSCR_OFFSET			(0x10) // OS Timer Count Register
#else // SA1110
//	#define PROCESSOR_BASE		0x90000000
//	#define GPIO_BASE			(PROCESSOR_BASE + 0x40000)
	#define OST_BASE			0x90000000  // OS Timers
	#define OSCR_OFFSET			(0x10) // OS Timer Count Register
#endif
#define	MEMORY_SIZE				0x10000  // 64kiB memory space; max allocation for VirtualAlloc is 64MiB

#define HNS_TO_TICKS(HNS)      ((ULONG)(((ULONGLONG)(HNS)*1583296744) >> 32))
// HNS_TO_TICKS converts the input in units of 100 ns to system ticks of
// a 3.6864 MHz clock.  It uses fixed point math operations with shifts
// instead of divides for speed.  The transformation is as follows.
//
// 3.6864 MHz gives a 271.267361111 ns period (one tick).  The conversion
// ratio is 100 ns = 1 hns = 0.36864 ticks.  To preserve precision under
// fixed point arithmetic, the conversion ratio is biased upward by a
// factor of 2^32.  This gives 0,36864 * 2^32 = 1583296744.  When this
// number is multiplied with the number of 100 ns intervals (hns), the
// product is the number of ticks * 2^32 for that interval.  Dividing
// out the extra factor of 2^32 in the product yields the correct number
// of ticks for the interval.  This is done with a right shift of 32.
//
// Note that the arithmetic is done with 64-bit integers to gain precision.
// The maximum number of 100 ns intervals this macro must handle is 2^32.
// Since the conversion factor is only a 31-bit number, the product can
// have a maximum of 63 bits.  This will not overflow a 64-bit integer.
//

// VirtualCopy definition
extern "C" BOOL VirtualCopy(LPVOID lpvDestMem, LPVOID lpvSrcMem, 
							DWORD dwSizeInBytes, DWORD dwProtectFlag);

PerfSleep::PerfSleep()
{
	pOSTCtrl = (DWORD *)MapAddress(OST_BASE);
}

PerfSleep::~PerfSleep()
{
	if(pOSTCtrl)
		ReleaseAddress(pOSTCtrl);
}

LPVOID PerfSleep::MapAddress(DWORD baseAddress)
{
	LPVOID pAddr = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
	
	if(pAddr)	
	{
		if(VirtualCopy (pAddr, (PVOID)(baseAddress >> 8), 
				MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
				DEBUGMSG(TRUE, (_T("0x%08x mapped to: 0x%08x\r\n"), baseAddress, pAddr));
		}
		else 
		{
			RETAILMSG(TRUE, (_T("Error! at VirtualCopy()\r\n")));

			return NULL;
		}
	}
	else 
	{
		RETAILMSG(TRUE, (_T("Error! at VirtualAlloc()\r\n")));
		return NULL;
	}

	return pAddr;
}

void PerfSleep::ReleaseAddress(LPVOID address)
{
	if(address != NULL)
	{
		VirtualFree(address, 0, MEM_RELEASE);
		address = NULL;
	}
}

void PerfSleep::StallExecution(int ticks)
{
    ULONG StallTicks = 0;
    ULONG StartOscr = 0;
    ULONG CurrentOscr = 0;
    LONG ElapsedOscr = 0;

    // Read the OSCR to find the current time.
	StartOscr = (ULONG)*(pOSTCtrl + OSCR_OFFSET);

    StallTicks = HNS_TO_TICKS(ticks); // 1 tick == 100ns
    
    // Wait for the time to elapse.
    do {
        // Read the OSCR.
		CurrentOscr = (ULONG)*(pOSTCtrl + OSCR_OFFSET);

        // Calculate the elapsed OSCR time, taking into account a possible
        // wrap from 0xFFFFFFFF to 0x00000000.
        ElapsedOscr = (LONG)CurrentOscr - (LONG)StartOscr;
    } while ((ULONG)ElapsedOscr < StallTicks);
}

